--[[
	Expose More Source Lib
	For exposing more things from the C source code via Lua implementations

	Changelog

	v0.1:
		Initial, made specifically for the DKR Self-Magnet xItem
]]

local FRACBITS = FRACBITS
local INT32_MAX = INT32_MAX
local INT32_MIN = INT32_MIN
local MFE_VERTICALFLIP = MFE_VERTICALFLIP

-- based on commit 684b016487823dbb2922b480455fd6dff78aedee
local VERSION = {0, 1}

local function isNewerVersion()
	-- check if library even exists yet
	if not ExposeMoreSource then
		return true
	end

	-- compare version numbers
	for versionIndex = 1, min(#VERSION, #ExposeMoreSource.VERSION) do
		if VERSION[versionIndex] < ExposeMoreSource.VERSION[versionIndex] then
			return false
		elseif VERSION[versionIndex] > ExposeMoreSource.VERSION[versionIndex]
			return true
		end
	end

	-- version numbers for available indices match, but there could be more in either one
	return #VERSION >= #ExposeMoreSource.VERSION
end

if not isNewerVersion() then
	return
end

if not ExposeMoreSource then
	rawset(_G, "ExposeMoreSource", {
		VERSION = VERSION,
		utility = {},
		-- DOUBLE UNDERSCORE MEANS KEEP OUT. USE THE GLOBALS
		__injectedNames = {},
		__constants = {},
		__functions = {}
	})
end

local exposeMoreSourcePrep = {
	constants = {
		BOXTOP = 0,
		BOXBOTTOM = 1,
		BOXLEFT = 2,
		BOXRIGHT = 3
	},
	functions = {},
	helpers = {},
	utility = {}
}

function exposeMoreSourcePrep.utility.exposeMoreLine_t(line)
	if not (line and line.valid and pcall(function() local _ = line.frontside end)) then
		error("This isn't a valid line_t")
	end

	local lineWrapper = {}
	setmetatable(lineWrapper, {
		__index = function(wrapper, key)
			if key == "bbox" then
				return {
					[BOXTOP] = max(line.v1.y, line.v2.y),
					[BOXBOTTOM] = min(line.v1.y, line.v2.y),
					[BOXLEFT] = min(line.v1.x, line.v2.x),
					[BOXRIGHT] = max(line.v1.x, line.v2.x)
				}
			end
			return line[key]
		end,
		__len = function(wrapper)
			return #line
		end
	})
	return lineWrapper
end

-- convenience function to simulate I_Error
function exposeMoreSourcePrep.helpers.LI_Error(errorFormat, ...)
	error(string.format("\133LI_Error():\128 %s", string.format(errorFormat, ...)))
end

-- convenience function to simulate I_Assert
function exposeMoreSourcePrep.helpers.LI_Assert(assertion, assertionStringized, callerFunction)
	if assertion then
		return
	end
	ExposeMoreSource.__functions.LI_Error("assert failed: %s, function %s", assertionStringized, callerFunction)
end

function exposeMoreSourcePrep.functions.P_PointOnLineSide(pointX, pointY, line)
	local lineVertex1 = line.v1
	local differenceX, differenceY, leftCheck, rightCheck

	if not line.dx then
		if pointX <= lineVertex1.x then
			return line.dy > 0 and 1 or 0
		end

		return line.dy < 0 and 1 or 0
	end
	if not line.dy then
		if pointY <= lineVertex1.y then
			return line.dx < 0 and 1 or 0
		end

		return line.dx < 0 and 1 or 0
	end

	differenceX = pointX - lineVertex1.x
	differenceY = pointY - lineVertex1.y

	leftCheck = FixedMul(line.dy >> FRACBITS, differenceX)
	rightCheck = FixedMul(differenceY, line.dx >> FRACBITS)

	if rightCheck < leftCheck then
		return 0
	end
	return 1
end

function exposeMoreSourcePrep.functions.P_BoxOnLineSide(boundingBox, line)
	local test1, test2

	if line.slopetype == "horizontal" then
		test1 = boundingBox[BOXTOP] > line.v1.y and 1 or 0
		test2 = boundingBox[BOXBOTTOM] > line.v1.y and 1 or 0
		if line.dx < 0 then
			test1 = $ ^ 1
			test2 = $ ^ 1
		end
	elseif line.slopetype == "vertical" then
		test1 = boundingBox[BOXRIGHT] < line.v1.x and 1 or 0
		test2 = boundingBox[BOXLEFT] < line.v1.x and 1 or 0
		if line.dy < 0 then
			test1 = $ ^ 1
			test2 = $ ^ 1
		end
	elseif line.slopetype == "positive" then
		test1 = ExposeMoreSource.__functions.P_PointOnLineSide(boundingBox[BOXLEFT], boundingBox[BOXTOP], line)
		test2 = ExposeMoreSource.__functions.P_PointOnLineSide(boundingBox[BOXRIGHT], boundingBox[BOXBOTTOM], line)
	elseif line.slopetype == "negative" then
		test1 = ExposeMoreSource.__functions.P_PointOnLineSide(boundingBox[BOXRIGHT], boundingBox[BOXTOP], line)
		test2 = ExposeMoreSource.__functions.P_PointOnLineSide(boundingBox[BOXLEFT], boundingBox[BOXBOTTOM], line)
	else
		ExposeMoreSource.__functions.LI_Error("P_BoxOnLineSide: unknown slopetype %s", line.slopetype)
		return -1
	end

	if test1 == test2 then
		return test1
	end
	return -1
end

function exposeMoreSourcePrep.functions.HighestOnLine(checkRadius, checkX, checkY, checkLine, slope, insteadGetLowest)
	local lineVertex1Copy, lineVertex2Copy = {}, {}
	lineVertex1Copy.x = checkLine.v1.x
	lineVertex1Copy.y = checkLine.v1.y
	lineVertex2Copy.x = checkLine.v2.x
	lineVertex2Copy.y = checkLine.v2.y

	if abs(lineVertex1Copy.x - checkX) > checkRadius then
		local checkDifference = abs(lineVertex1Copy.x - checkX) - checkRadius

		if lineVertex1Copy.x < checkX then
			lineVertex1Copy.x = $ + checkDifference
			lineVertex1Copy.y = $ + FixedMul(checkDifference, FixedDiv(checkLine.dy, checkLine.dx))
		else
			lineVertex1Copy.x = $ - checkDifference
			lineVertex1Copy.y = $ - FixedMul(checkDifference, FixedDiv(checkLine.dy, checkLine.dx))
		end
	end

	if abs(lineVertex1Copy.y - checkY) > checkRadius then
		local checkDifference = abs(lineVertex1Copy.y - checkY) - checkRadius

		if lineVertex1Copy.y < checkY then
			lineVertex1Copy.y = $ + checkDifference
			lineVertex1Copy.x = $ + FixedMul(checkDifference, FixedDiv(checkLine.dx, checkLine.dy))
		else
			lineVertex1Copy.y = $ - checkDifference
			lineVertex1Copy.x = $ - FixedMul(checkDifference, FixedDiv(checkLine.dx, checkLine.dy))
		end
	end

	if abs(lineVertex2Copy.x - checkX) > checkRadius then
		local checkDifference = abs(lineVertex2Copy.x - checkX) - checkRadius

		if lineVertex2Copy.x < checkX then
			lineVertex2Copy.x = $ + checkDifference
			lineVertex2Copy.y = $ + FixedMul(checkDifference, FixedDiv(checkLine.dy, checkLine.dx))
		else
			lineVertex2Copy.x = $ - checkDifference
			lineVertex2Copy.y = $ - FixedMul(checkDifference, FixedDiv(checkLine.dy, checkLine.dx))
		end
	end

	if abs(lineVertex2Copy.y - checkY) > checkRadius then
		local checkDifference = abs(lineVertex2Copy.y - checkY) - checkRadius

		if lineVertex2Copy.y < checkY then
			lineVertex2Copy.y = $ + checkDifference
			lineVertex2Copy.x = $ + FixedMul(checkDifference, FixedDiv(checkLine.dx, checkLine.dy))
		else
			lineVertex2Copy.y = $ - checkDifference
			lineVertex2Copy.x = $ - FixedMul(checkDifference, FixedDiv(checkLine.dx, checkLine.dy))
		end
	end

	if insteadGetLowest then
		return min(
			P_GetZAt(slope, lineVertex1Copy.x, lineVertex1Copy.y),
			P_GetZAt(slope, lineVertex2Copy.x, lineVertex2Copy.y)
		)
	else
		return max(
			P_GetZAt(slope, lineVertex1Copy.x, lineVertex1Copy.y),
			P_GetZAt(slope, lineVertex2Copy.x, lineVertex2Copy.y)
		)
	end
end

function exposeMoreSourcePrep.functions.P_MobjCeilingZ(mobj, sourceSector, boundingSector, checkX, checkY, touchingLine, getLowest, bePerfect)
	ExposeMoreSource.__functions.LI_Assert(mobj and mobj.valid, "mobj and mobj.valid", "P_MobjCeilingZ")
	ExposeMoreSource.__functions.LI_Assert(sourceSector and sourceSector.valid, "sourceSector and sourceSector.valid", "P_MobjCeilingZ")
	if sourceSector.c_slope then
		local testX, testY
		local slope = sourceSector.c_slope

		if slope.d.x < 0 then
			testX = mobj.radius
		else
			testX = -mobj.radius
		end

		if slope.d.y < 0
			testY = mobj.radius
		else
			testY = -mobj.radius
		end

		if (slope.zdelta > 0 and 1 or 0) ^ (getLowest and 1 or 0) then
			testX = -testX
			testY = -testY
		end

		testX = $ + checkX
		testY = $ + checkY

		if R_PointInSubsector(testX, testY).sector == (boundingSector and boundingSector.valid and boundingSector or sourceSector) then
			return P_GetZAt(slope, testX, testY)
		end

		if bePerfect then
			local lineIndex
			local boundingLine, boundingLineExposed
			local boundingBox = {}
			local finalHeight

			if getLowest then
				finalHeight = INT32_MAX
			else
				finalHeight = INT32_MIN
			end

			boundingBox[BOXLEFT] = checkX - mobj.radius
			boundingBox[BOXRIGHT] = checkX + mobj.radius
			boundingBox[BOXTOP] = checkY + mobj.radius
			boundingBox[BOXBOTTOM] = checkY - mobj.radius
			for lineIndex = 0, #boundingSector.lines - 1 do
				boundingLine = boundingSector.lines[lineIndex]
                boundingLineExposed = ExposeMoreSource.__functions.exposeMoreLine_t(boundingLine)

				if boundingBox[BOXRIGHT] <= boundingLineExposed.bbox[BOXLEFT] or boundingBox[BOXLEFT] >= boundingLineExposed.bbox[BOXRIGHT]
				or boundingBox[BOXTOP] <= boundingLineExposed.bbox[BOXBOTTOM] or boundingBox[BOXBOTTOM] >= boundingLineExposed.bbox[BOXTOP] then
					continue
				end

				if ExposeMoreSource.__functions.P_BoxOnLineSide(boundingBox, boundingLine) ~= -1 then
					continue
				end

				if getLowest then
					finalHeight = min(finalHeight, ExposeMoreSource.__functions.HighestOnLine(mobj.radius, checkX, checkY, boundingLine, slope, true))
				else
					finalHeight = max(finalHeight, ExposeMoreSource.__functions.HighestOnLine(mobj.radius, checkX, checkY, boundingLine, slope, false))
				end
			end

			return finalHeight
		end

		if not (touchingLine and touchingLine.valid) then
			return P_GetZAt(slope, checkX, checkY)
		end

		return ExposeMoreSource.__functions.HighestOnLine(mobj.radius, checkX, checkY, touchingLine, slope, getLowest)
	else
		return sourceSector.ceilingheight
	end
end

function exposeMoreSourcePrep.functions.P_GetSpecialTopZ(mobj, sourceSector, boundingSector)
	return ExposeMoreSource.__functions.P_MobjCeilingZ(mobj, sourceSector, boundingSector, mobj.x, mobj.y, nil, sourceSector == boundingSector, true)
end

function exposeMoreSourcePrep.functions.P_MobjFloorZ(mobj, sourceSector, boundingSector, checkX, checkY, touchingLine, getLowest, bePerfect)
	ExposeMoreSource.__functions.LI_Assert(mobj and mobj.valid, "mobj and mobj.valid", "P_MobjFloorZ")
	ExposeMoreSource.__functions.LI_Assert(sourceSector and sourceSector.valid, "sourceSector and sourceSector.valid", "P_MobjFloorZ")
	if sourceSector.f_slope then
		local testX, testY
		local slope = sourceSector.f_slope

		if slope.d.x < 0 then
			testX = mobj.radius
		else
			testX = -mobj.radius
		end

		if slope.d.y < 0
			testY = mobj.radius
		else
			testY = -mobj.radius
		end

		if (slope.zdelta > 0 and 1 or 0) ^ (getLowest and 1 or 0) then
			testX = -testX
			testY = -testY
		end

		testX = $ + checkX
		testY = $ + checkY

		if R_PointInSubsector(testX, testY).sector == (boundingSector and boundingSector.valid and boundingSector or sourceSector) then
			return P_GetZAt(slope, testX, testY)
		end

		if bePerfect then
			local lineIndex
			local boundingLine, boundingLineExposed
			local boundingBox = {}
			local finalHeight

			if getLowest then
				finalHeight = INT32_MAX
			else
				finalHeight = INT32_MIN
			end

			boundingBox[BOXLEFT] = checkX - mobj.radius
			boundingBox[BOXRIGHT] = checkX + mobj.radius
			boundingBox[BOXTOP] = checkY + mobj.radius
			boundingBox[BOXBOTTOM] = checkY - mobj.radius
			for lineIndex = 0, #boundingSector.lines - 1 do
				boundingLine = boundingSector.lines[lineIndex]
                boundingLineExposed = ExposeMoreSource.__functions.exposeMoreLine_t(boundingLine)

				if boundingBox[BOXRIGHT] <= boundingLineExposed.bbox[BOXLEFT] or boundingBox[BOXLEFT] >= boundingLineExposed.bbox[BOXRIGHT]
				or boundingBox[BOXTOP] <= boundingLineExposed.bbox[BOXBOTTOM] or boundingBox[BOXBOTTOM] >= boundingLineExposed.bbox[BOXTOP] then
					continue
				end

				if ExposeMoreSource.__functions.P_BoxOnLineSide(boundingBox, boundingLine) ~= -1 then
					continue
				end

				if getLowest then
					finalHeight = min(finalHeight, ExposeMoreSource.__functions.HighestOnLine(mobj.radius, checkX, checkY, boundingLine, slope, true))
				else
					finalHeight = max(finalHeight, ExposeMoreSource.__functions.HighestOnLine(mobj.radius, checkX, checkY, boundingLine, slope, false))
				end
			end

			return finalHeight
		end

		if not (touchingLine and touchingLine.valid) then
			return P_GetZAt(slope, checkX, checkY)
		end

		return ExposeMoreSource.__functions.HighestOnLine(mobj.radius, checkX, checkY, touchingLine, slope, getLowest)
	else
		return sourceSector.floorheight
	end
end

function exposeMoreSourcePrep.functions.P_GetSpecialBottomZ(mobj, sourceSector, boundingSector)
	return ExposeMoreSource.__functions.P_MobjFloorZ(mobj, sourceSector, boundingSector, mobj.x, mobj.y, nil, sourceSector ~= boundingSector, true)
end

function exposeMoreSourcePrep.functions.P_IsObjectOnRealGround(mobj, sector)
	if mobj.eflags & MFE_VERTICALFLIP then
		if mobj.z + mobj.height >= ExposeMoreSource.__functions.P_GetSpecialTopZ(mobj, sector, sector) then
			return true
		end
	else
		if mobj.z <= ExposeMoreSource.__functions.P_GetSpecialBottomZ(mobj, sector, sector) then
			return true
		end
	end
	return false
end

local function shouldInjectLuaImplementation(targetName)
	-- only inject if the global doesn't already exist, or it was replaced by the older lib
	return not _G[targetName] or ExposeMoreSource.__injectedNames[targetName]
end

-- don't clobber constants/functions possibly added in future
for constantName, constantValue in pairs(exposeMoreSourcePrep.constants) do
	if shouldInjectLuaImplementation(constantName) then
		ExposeMoreSource.__injectedNames[constantName] = true
		rawset(_G, constantName, constantValue)
	end
end

for functionName, functionDefinition in pairs(exposeMoreSourcePrep.functions) do
	if shouldInjectLuaImplementation(functionName) then
		ExposeMoreSource.__functions[functionName] = functionDefinition
		rawset(_G, functionName, function(...)
			local result = {pcall(functionDefinition, ...)}
			local ok = table.remove(result, 1)
			if ok then
				return unpack(result)
			else
				-- if something goes wrong intentionally, blame the caller
				local errorMessage = result[1]
				error(errorMessage, 2)
			end
		end)
		-- must be separate since lib function could simply wrap source
		ExposeMoreSource.__injectedNames[functionName] = true
	else
		-- use the actual global for backdoor calls
		ExposeMoreSource.__functions[functionName] = _G[functionName]
	end
end

for helperName, helperDefinition in pairs(exposeMoreSourcePrep.helpers) do
	ExposeMoreSource.__functions[helperName] = helperDefinition
end

for utilityName, utilityDefinition in pairs(exposeMoreSourcePrep.utility) do
	ExposeMoreSource.__functions[utilityName] = utilityDefinition
	ExposeMoreSource.utility[utilityName] = function(...)
		local result = {pcall(utilityDefinition, ...)}
		local ok = table.remove(result, 1)
		if ok then
			return unpack(result)
		else
			-- if something goes wrong intentionally, blame the caller
			local errorMessage = result[1]
			error(errorMessage, 2)
		end
	end
end

ExposeMoreSource.VERSION = VERSION
